// MD2Model.cpp
#include <windows.h>
#include <d3d9.h>
#include <fstream>
#include "SafeRelease.h"

#include "MD2Model.h"

// CMD2Model
int CMD2Model::LoadFromFile(char* file_name, IDirect3DDevice9* device)
{
	try
	{
		// Open the file
		std::ifstream md2_file;
		md2_file.open(file_name, std::ios::in | std::ios::binary);

		// Check if file has been opened
		if (!md2_file.is_open())
			throw MD2LoadException("Could not open model file");

		// Load file
		return Load(md2_file, device);
	}
	catch (MD2LoadException& e)
	{
		MessageBox(0,e.error,"MD2LoadException",0);
	}
	return 1;
}

int CMD2Model::Load(std::istream& md2_stream, IDirect3DDevice9* device)
{
	try
	{
		is_valid = false;
		// Get and validate header
		GetHeader(md2_stream);
		ValidateHeader();

		GetFrameData(md2_stream, device);
		GetIndexData(md2_stream, device);
		GetTextureData(md2_stream, device);

		is_valid = true;
		return 0;
	}
	catch (MD2LoadException& e)
	{
		MessageBox(0,e.error,"MD2LoadException",0);
	}
	return 1;
}

void CMD2Model::GetHeader(std::istream& md2_stream)
{
	md2_stream.seekg(0, std::ios::beg);
	md2_stream.read((char*)&header, sizeof(md2::header));
}


void CMD2Model::GetFrameData(std::istream& md2_stream, IDirect3DDevice9* device)
{
	ValidateHeader();
	
	int frame_data_size = (header.framesize * header.num_frames);
	char* buffer = new char[frame_data_size];
	md2_stream.seekg(header.ofs_frames, std::ios::beg);
	md2_stream.read((char*)buffer, frame_data_size);

	MD2_POSITION_VERTEX* vertices = new MD2_POSITION_VERTEX[header.num_xyz];

	md2::frame* cur_frame = (md2::frame*)buffer;
	for (int f = 0; f < header.num_frames; f++)
	{
		cur_frame = (md2::frame*) &buffer[header.framesize * f];
		for (int v = 0; v < header.num_xyz; v++)
		{
			for (int i = 0; i < 3; i++)
				vertices[v].a[i] = (cur_frame->verticies[v].vertex[i] * cur_frame->scale[i])
										+cur_frame->translate[i];
		}
		bool is_created = D3D_OK == device->CreateVertexBuffer(sizeof(MD2_POSITION_VERTEX)*header.num_xyz,0,0,D3DPOOL_MANAGED,&model_VB[f],0);
		if (is_created == 0)
		{
			MessageBox(0,"Could not create vertex buffers", "Fatal Error", 0);
			PostQuitMessage(0);
			return;
		}
		void* data_ptr = 0;
		bool is_locked = D3D_OK == model_VB[f]->Lock(0, 0, &data_ptr, 0);//D3DLOCK_DISCARD
		memcpy(data_ptr, vertices, sizeof(MD2_POSITION_VERTEX)*header.num_xyz);
		model_VB[f]->Unlock();

		used_VBs++;
	}

	delete[] vertices;
	delete[] buffer;
}

void CMD2Model::GetIndexData(std::istream& md2_stream, IDirect3DDevice9* device)
{	
	ValidateHeader();

	md2::triangle* tri_buffer = new md2::triangle[header.num_tris];
	md2::index_buffer* index_buffer = new md2::index_buffer[header.num_tris];

	md2_stream.seekg(header.ofs_tris, std::ios::beg);
	md2_stream.read((char*)tri_buffer, sizeof(md2::triangle)*header.num_tris);

	for (int i = 0; i < header.num_tris; i++)
	{
		index_buffer[i][0] = tri_buffer[i].vertex_indicies[0];
		index_buffer[i][1] = tri_buffer[i].vertex_indicies[1];
		index_buffer[i][2] = tri_buffer[i].vertex_indicies[2];
	}

	device->CreateIndexBuffer(sizeof(md2::index_buffer)*header.num_tris,0,D3DFMT_INDEX16,D3DPOOL_MANAGED,&model_IB,0);

	void* data_ptr;
	model_IB->Lock(0, 0, &data_ptr, 0);
	memcpy(data_ptr, index_buffer, sizeof(md2::index_buffer)*header.num_tris);
	model_IB->Unlock();

	delete[] index_buffer;
	delete[] tri_buffer;
}

void CMD2Model::GetTextureData(std::istream& md2_stream, IDirect3DDevice9* device)
{
	ValidateHeader();

	md2::tex_coord* c_buffer = new md2::tex_coord[header.num_st];
	MD2_TEXTURE_VERTEX* tex_buffer = new MD2_TEXTURE_VERTEX[header.num_st];

	md2_stream.seekg(header.ofs_st, std::ios::beg);
	md2_stream.read((char*)c_buffer, sizeof(md2::tex_coord)*header.num_st);

	for (int i = 0; i < header.num_st; i++)
	{
		tex_buffer[i].s = ((float)c_buffer[i].s / (float)(header.skinwidth));
		tex_buffer[i].t = ((float)c_buffer[i].t / (float)(header.skinheight));
	}

	// get the tex coord ref'd in the indices, then move it into vert_tex_buffer
	MD2_TEXTURE_VERTEX* vert_tex_buffer = new MD2_TEXTURE_VERTEX[header.num_xyz];
	md2::triangle* tri_buffer = new md2::triangle[header.num_tris];

	md2_stream.seekg(header.ofs_tris, std::ios::beg);
	md2_stream.read((char*)tri_buffer, sizeof(md2::triangle)*header.num_tris);

	for (int t = 0; t < header.num_tris; t++)
	{
		for (i = 0; i < 3; i++)
		{
			int vertex_i = tri_buffer[t].vertex_indicies[i];
			vert_tex_buffer[vertex_i].s = tex_buffer[tri_buffer[t].texture_indicies[i]].s;
			vert_tex_buffer[vertex_i].t = tex_buffer[tri_buffer[t].texture_indicies[i]].t;
		}
	}

	device->CreateVertexBuffer(sizeof(MD2_TEXTURE_VERTEX)*header.num_xyz,0,0,D3DPOOL_MANAGED,&texture_VB,0);
		
	void* data_ptr;
	texture_VB->Lock(0, 0, &data_ptr, 0);
	memcpy(data_ptr, vert_tex_buffer, sizeof(MD2_TEXTURE_VERTEX)*header.num_xyz);
	texture_VB->Unlock();

	delete[] tri_buffer;
	delete[] vert_tex_buffer;
	delete[] tex_buffer;
	delete[] c_buffer;
}

void CMD2Model::ValidateHeader(void)
{
	if (header.magic_number != 844121161 || header.version != 8 || header.framesize == 0)
		throw MD2LoadException("Invalid md2 header");;
}

IDirect3DVertexBuffer9* CMD2Model::GetModelVB(int frame)
{
	return model_VB[frame];
}

IDirect3DVertexBuffer9* CMD2Model::GetTextureVB(void)
{
	return texture_VB;
}

IDirect3DIndexBuffer9* CMD2Model::GetModelIB(void)
{
	return model_IB;
}

int CMD2Model::GetVertCount(void)
{
	return header.num_xyz;
}

int CMD2Model::GetTriangleCount(void)
{
	return header.num_tris;
}

int CMD2Model::GetFrameCount(void)
{
	return header.num_frames;
}

char* CMD2Model::GetFrameName(int frame)
{
	return 0;
}

CMD2Model::CMD2Model(void)
{
	is_valid = false;

	used_VBs = 0;
	// Zero structs
	memset(&header,0,sizeof(md2::header));

	// Zero pointers
	for (int i = 0; i < 512; i++)
		model_VB[i] = 0;

	texture_VB = 0;
	model_IB = 0;
}

CMD2Model::~CMD2Model(void)
{
	for (int i = 0; i < used_VBs; i++)
		SafeRelease(model_VB[i]);

	SafeRelease(texture_VB);
	SafeRelease(model_IB);
}
